from fmtr.tools.import_tools import MissingExtraMockModule
from fmtr.tools.path_tools.base import Path, PackagePaths

try:
    from fmtr.tools.path_tools.app import AppPaths
except ImportError as exception:
    AppPaths = MissingExtraMockModule('path.app', exception)



from pydantic import BaseModel, RootModel, ConfigDict


def to_df(*objs, name_value='value'):
    """

    DataFrame representation of Data Models as rows.

    """
    from fmtr.tools import tabular

    rows = []
    for obj in objs:
        if isinstance(obj, BaseModel):
            row = obj.model_dump()
        else:
            row = {name_value: obj}
        rows.append(row)

    df = tabular.DataFrame(rows)
    if 'id' in df.columns:
        df.set_index('id', inplace=True, drop=True)
    return df


class MixinArbitraryTypes:
    """

    Convenience for when non-serializable types are needed
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

class MixinFromJson:

    @classmethod
    def from_json(cls, json_str):
        """

        Error-tolerant deserialization

        """
        from fmtr.tools import json_fix
        data = json_fix.from_json(json_str, default={})

        if type(data) is dict:
            self = cls(**data)
        else:
            self = cls(data)

        return self


class Base(BaseModel, MixinFromJson):
    """

    Base model

    """

    def to_df(self, name_value='value'):
        """

        DataFrame representation with Fields as rows.

        """

        objs = []
        for name in self.model_fields.keys():
            val = getattr(self, name)
            objs.append(val)

        df = to_df(*objs, name_value=name_value)
        df['id'] = list(self.model_fields.keys())
        df = df.set_index('id', drop=True)
        return df


class Root(RootModel, MixinFromJson):
    """

    Root (list) model

    """

    def to_df(self):
        """

        DataFrame representation with items as rows.

        """

        return to_df(*self.items)



from typing import List, Optional

import pydantic_ai
from pydantic_ai import RunContext
from pydantic_ai.agent import AgentRunResult, Agent
from pydantic_ai.models.openai import OpenAIModel
from pydantic_ai.providers.openai import OpenAIProvider

from fmtr.tools import environment_tools as env
from fmtr.tools.constants import Constants
from fmtr.tools.string_tools import truncate_mid

pydantic_ai.Agent.instrument_all()

class Task:
    """

    Linear task definition, as Agent configuration and typing, plus state history.

    """

    PROVIDER = OpenAIProvider(api_key=env.get(Constants.FMTR_OPENAI_API_KEY_KEY))

    API_HOST_FMTR = env.get(Constants.FMTR_AI_HOST_KEY, Constants.FMTR_AI_HOST_DEFAULT)
    API_URL_FMTR = f'https://{API_HOST_FMTR}/v1'
    PROVIDER_FMTR = OpenAIProvider(base_url=API_URL_FMTR)

    MODEL_ID = 'gpt-4o'
    MODEL_ID_FMTR = 'qwen2.5-coder:14b'
    SYSTEM_PROMPT_STATIC = None
    DEPS_TYPE = str
    RESULT_TYPE = str
    RESULT_RETRIES = 5

    def __init__(self, *args, **kwargs):
        """

        Configure Agent

        """

        self.model = OpenAIModel(self.MODEL_ID, provider=self.PROVIDER)
        self.agent = Agent(
            *args,
            model=self.model,
            system_prompt=self.SYSTEM_PROMPT_STATIC or [],
            deps_type=self.DEPS_TYPE,
            result_type=self.RESULT_TYPE,
            result_retries=self.RESULT_RETRIES,
            **kwargs
        )

        self.agent.output_validator(self.validate)
        self.agent.system_prompt(self.add_system_prompt)
        self.history = []

    @property
    def sync_runner(self):
        """

        Convenience/debug function to run without async.

        """
        import asyncio
        return asyncio.run

    async def run(self, *args, deps=None, **kwargs) -> AgentRunResult[RESULT_TYPE]:
        """

        Run Agent with deps-relative user prompt and while storing history

        """
        result = await self.agent.run(*args, user_prompt=self.get_prompt(deps), deps=deps, message_history=self.history, **kwargs)
        self.history = result.all_messages()
        return result

    async def validate(self, ctx: RunContext[DEPS_TYPE], output: RESULT_TYPE) -> RESULT_TYPE:
        """

        Dummy validator

        """
        return output

    def get_prompt(self, deps: Optional[DEPS_TYPE]) -> Optional[str]:
        """

        Dummy prompt generator

        """
        return None

    def add_system_prompt(self, ctx: RunContext[DEPS_TYPE]) -> str | List[str]:
        """

        Dummy system prompt append

        """

        return []

    def __repr__(self):
        """

        String representation of the object

        """
        return f'{self.__class__.__name__}({repr(truncate_mid(self.SYSTEM_PROMPT_STATIC, 100))})'

if __name__ == '__main__':
    import asyncio
    from fmtr.tools import dm


    class TestOutput(dm.Base):
        text: str


    class TestDeps(dm.Base):
        lang: str
        subject: str


    class TaskTest(Task):
        # PROVIDER = Task.PROVIDER_FMTR
        # MODEL_ID = 'qwen2.5-coder:14b'
        RESULT_TYPE = TestOutput
        SYSTEM_PROMPT_STATIC = 'Tell the user jokes.'

        def add_system_prompt(self, ctx: RunContext[TestDeps]) -> str:
            return f'The jokes must be in the {ctx.deps.lang} language.'

        def get_prompt(self, deps: Optional[TestDeps]) -> str:
            return f'Tell me one about {deps.subject}.'

    task = TaskTest()
    deps = TestDeps(lang='English', subject='eggs')
    result1 = task.sync_runner(task.run(deps=deps))
    result1

    deps = TestDeps(lang='German', subject='sausages')
    result2 = task.sync_runner(task.run(deps=deps))
    result2



import json
import json_repair

from fmtr.tools.logging_tools import logger
from fmtr.tools.tools import Raise


def from_json(json_string, default=None):
    """

    Error-tolerant JSON deserialization

    """
    try:
        return json_repair.loads(json_string)
    except json.JSONDecodeError as exception:
        if default is Raise:
            raise exception
        logger.warning(f'Deserialization failed {repr(exception)}: {json_string}')
        return default



import re
import subprocess
from pathlib import Path
from tempfile import gettempdir
from typing import Union, Any

from fmtr.tools.constants import Constants
from fmtr.tools.platform_tools import is_wsl

WIN_PATH_PATTERN = r'''([a-z]:(\\|$)|\\\\)'''
WIN_PATH_RX = re.compile(WIN_PATH_PATTERN, flags=re.IGNORECASE)


class WSLPathConversionError(EnvironmentError):
    """

    Error to raise if WSL path conversion fails.

    """


class Path(type(Path())):
    """

    Custom path object aware of WSL paths, with some additional read/write methods

    """

    def __new__(cls, *segments: Union[str, Path], convert_wsl: bool = True, **kwargs):
        """

        Intercept arguments to detect whether WSL conversion is required.

        """
        if convert_wsl and len(segments) == 1 and is_wsl() and cls.is_abs_win_path(*segments):
            segments = [cls.from_wsl(*segments)]

        return super().__new__(cls, *segments, **kwargs)

    @classmethod
    def is_abs_win_path(cls, path: Union[str, Path]) -> bool:
        """

        Infer if the current path is an absolute Windows path.

        """
        path = str(path)
        return bool(WIN_PATH_RX.match(path))

    @classmethod
    def from_wsl(cls, path: Union[str, Path]) -> bool:  # pragma: no cover
        """

        Call `wslpath` to convert the path to its Unix equivalent.

        """
        result = subprocess.run(['wslpath', '-u', str(path)], capture_output=True, text=True)

        if result.returncode:
            msg = f'Could not convert Windows path to Unix equivalent: "{path}"'
            raise WSLPathConversionError(msg)

        path_wsl = result.stdout.strip()
        path_wsl = cls(path_wsl, convert_wsl=False)
        return path_wsl

    @classmethod
    def package(cls) -> 'Path':
        """

        Get path to originating module (e.g. directory containing .py file).

        """
        from fmtr.tools.inspection_tools import get_call_path
        path = get_call_path(offset=2).absolute().parent
        return path

    @classmethod
    def module(cls) -> 'Path':
        """

        Get path to originating module (i.e. .py file).

        """
        from fmtr.tools.inspection_tools import get_call_path
        path = get_call_path(offset=2).absolute()
        return path

    @classmethod
    def temp(cls) -> 'Path':
        """

        Get path to temporary directory.

        """
        return cls(gettempdir())

    def write_json(self, obj) -> int:
        """

        Write the specified object to the path as a JSON string

        """
        from fmtr.tools import json
        json_str = json.to_json(obj)
        return self.write_text(json_str, encoding=Constants.ENCODING)

    def read_json(self) -> Any:
        """

        Read JSON from the file and return as a Python object

        """
        from fmtr.tools import json
        json_str = self.read_text(encoding=Constants.ENCODING)
        obj = json.from_json(json_str)
        return obj

    def write_yaml(self, obj) -> int:
        """

        Write the specified object to the path as a JSON string

        """
        from fmtr.tools import yaml
        yaml_str = yaml.to_yaml(obj)
        return self.write_text(yaml_str, encoding=Constants.ENCODING)

    def read_yaml(self) -> Any:
        """

        Read YAML from the file and return as a Python object

        """
        from fmtr.tools import yaml
        yaml_str = self.read_text(encoding=Constants.ENCODING)
        obj = yaml.from_yaml(yaml_str)
        return obj

    def mkdirf(self):
        """

        Convenience method for creating directory with parents

        """
        return self.mkdir(parents=True, exist_ok=True)

    @property
    def exist(self):
        """

        Exists as property

        """
        return super().exists()

    @classmethod
    def app(cls):
        """

        Convenience method for getting application paths

        """
        from fmtr.tools import path
        return path.AppPaths()

class PackagePaths:
    """

    Canonical paths for a package.

    """

    dev = Path('/') / 'opt' / 'dev'
    data_global = dev / Constants.DIR_NAME_DATA

    def __init__(self, path=None, org_singleton=None, dir_name_data=Constants.DIR_NAME_DATA, filename_config=Constants.FILENAME_CONFIG, file_version=Constants.FILENAME_VERSION):

        """

        Use calling module path as default path, if not otherwise specified.

        """
        if not path:
            from fmtr.tools.inspection_tools import get_call_path
            path = get_call_path(offset=2).parent

        self.path = Path(path)
        self.org_singleton = org_singleton
        self.dir_name_data = dir_name_data
        self.filename_config = filename_config
        self.filename_version = file_version

    @property
    def is_dev(self) -> bool:
        """

        Is the package in the dev directory - as opposed to `site-packages` etc?

        """
        return self.path.is_relative_to(self.dev)

    @property
    def is_namespace(self) -> bool:
        """

        If organization is not hard-specified, then the package is a namespace.

        """
        return not bool(self.org_singleton)

    @property
    def name(self) -> str:
        """

        Name of package.

        """
        return self.path.stem

    @property
    def name_ns(self) -> str:
        """

        Name of namespace package.

        """

        if self.is_namespace:
            return f'{self.org}.{self.name}'
        else:
            return self.name

    @property
    def org(self) -> str:
        """

        Name of organization.

        """
        if not self.is_namespace:
            return self.org_singleton
        else:
            return self.path.parent.stem

    @property
    def repo(self) -> Path:
        """

        Path of repo (i.e. parent of package base directory).

        """
        if self.is_namespace:
            return self.path.parent.parent
        else:
            return self.path.parent

    @property
    def version(self) -> Path:
        """

        Path of version file.

        """
        return self.path / self.filename_version

    @property
    def data(self) -> Path:
        """

        Path of project-specific data directory.

        """

        return self.dev / Constants.DIR_NAME_REPO / self.name_ns / self.dir_name_data

    @property
    def artifact(self) -> Path:
        """

        Path of project-specific artifact directory

        """

        return self.data / Constants.DIR_NAME_ARTIFACT

    @property
    def source(self) -> Path:
        """

        Path of project-specific source directory

        """

        return self.data / Constants.DIR_NAME_SOURCE



    @property
    def settings(self) -> Path:
        """

        Path of settings file.

        """
        return self.data / self.filename_config

    @property
    def hf(self) -> Path:
        """

        Path of HuggingFace directory

        """
        return self.artifact / Constants.DIR_NAME_HF

    def __repr__(self) -> str:
        """

        Show base path in repr.

        """
        return f'{self.__class__.__name__}("{self.path}")'


if __name__ == "__main__":
    path = Path()
    path.app.user_log_dir()
    path



from fmtr.tools.import_tools import MissingExtraMockModule

try:
    from fmtr.tools.ai_tools import inference_tools as infer
except ImportError as exception:
    infer = MissingExtraMockModule('ai', exception)

try:
    from fmtr.tools.ai_tools import agentic_tools as agentic
except ImportError as exception:
    agentic = MissingExtraMockModule('ai.client', exception)



import appdirs

from fmtr.tools.path_tools import Path


class AppPaths:
    """

    Wrap appdirs to return Path objects

    """
    PathType = Path

    def user_data_dir(self, appname=None, appauthor=None, version=None, roaming=False):
        path_str = appdirs.user_data_dir(appname=appname, appauthor=appauthor, version=version, roaming=roaming)
        return self.PathType(path_str)

    def user_config_dir(self, appname=None, appauthor=None, version=None, roaming=False):
        path_str = appdirs.user_config_dir(appname=appname, appauthor=appauthor, version=version, roaming=roaming)
        return self.PathType(path_str)

    def site_config_dir(self, appname=None, appauthor=None, version=None):
        path_str = appdirs.site_config_dir(appname=appname, appauthor=appauthor, version=version, multipath=False)
        return self.PathType(path_str)

    def site_data_dir(self, appname=None, appauthor=None, version=None):
        path_str = appdirs.site_data_dir(appname=appname, appauthor=appauthor, version=version, multipath=False)
        return self.PathType(path_str)

    def user_cache_dir(self, appname=None, appauthor=None, version=None):
        path_str = appdirs.user_cache_dir(appname=appname, appauthor=appauthor, version=version)
        return self.PathType(path_str)

    def user_state_dir(self, appname=None, appauthor=None, version=None):
        path_str = appdirs.user_state_dir(appname=appname, appauthor=appauthor, version=version)
        return self.PathType(path_str)

    def user_log_dir(self, appname=None, appauthor=None, version=None):
        path_str = appdirs.user_log_dir(appname=appname, appauthor=appauthor, version=version)
        return self.PathType(path_str)



from fmtr.tools import json_tools
from fmtr.tools.tests.helpers import SERIALIZATION_DATA


def test_json():
    """

    Simple YAML round trip test

    """
    expected = SERIALIZATION_DATA
    actual = json_tools.from_json(json_tools.to_json(expected))
    assert actual == expected



from multiprocessing import cpu_count

import dask
import dask.bag as db
from contextlib import nullcontext as NullContext
from dask.diagnostics import ProgressBar
from typing import List, Callable, Any, Union

from fmtr.tools.iterator_tools import dict_records_to_lists
from fmtr.tools.logging_tools import logger
from fmtr.tools.path_tools import Path


class ParallelConfig:
    """

    Configuration values.

    """
    THREADS = 'threads'
    PROCESSES = 'processes'
    SINGLE = 'single-threaded'


def get_nullary_wrapper(func: Callable):
    """

    Dask cannot map a nullary function, as its argument list is empty. Hence this wrapper to force the function to take one dummy argument.

    """

    def wrap_nullary(dummy: None, **kwargs):
        """

        Ignore the dummy argument and run the function.

        """
        return func(**kwargs)

    return wrap_nullary


def apply(func: Callable, data: Union[List[Any], int], *args, num_workers: int = cpu_count(),
          scheduler: str = ParallelConfig.PROCESSES,
          parallelize: bool = True, show_progress: bool = False, return_future: bool = False, **kwargs) -> \
        List[Any]:
    """

    Helper function for a one-off, intensive parallel computation task.

    """

    if not parallelize and scheduler != ParallelConfig.SINGLE:
        msg = f'Scheduler is set to "{scheduler}" but parallelization has been manually disabled.'
        logger.warning(msg)
        scheduler = ParallelConfig.SINGLE

    data_kwargs = {}
    if type(data) is int:  # If data is an integer, assume the function is nullary and just run it the specified number of times.
        data_args = [[None] * data]
        func = get_nullary_wrapper(func)
    else:
        data_args = []
        is_data_lists = all(isinstance(datum, (tuple, list)) for datum in data)
        is_data_dicts = all(isinstance(datum, dict) for datum in data)
        if is_data_lists:  # If the data is a list of tuples/lists of arguments.
            data_args += list(zip(*data))
        elif is_data_dicts:  # If the data is a list of dictionaries of keyword arguments.
            data_kwargs = dict_records_to_lists(data)
        else:
            data_args.append(data)  # Otherwise treat the data as a simple list of arguments.

    dask.config.set({'temporary-directory': Path.temp()})

    data_args = [db.from_sequence(value) for value in data_args]
    data_kwargs = {key: db.from_sequence(values) for key, values in data_kwargs.items()}
    future = db.map(func, *data_args, *args, **data_kwargs, **kwargs)

    def get_results():
        """

        Function to compute results with the specified configuration.

        """
        if show_progress:
            context = ProgressBar
        else:
            context = NullContext

        with context():
            return future.compute(scheduler=scheduler, num_workers=num_workers)

    if return_future:  # Return a delayed function.
        return get_results
    else:
        results = get_results()  # Compute and return results.
        return results



import time

import argparse
import huggingface_hub
import json
import os
from huggingface_hub.constants import HUGGINGFACE_HUB_CACHE
from huggingface_hub.file_download import repo_folder_name
from pathlib import Path

FUNCTIONS = [huggingface_hub.snapshot_download]
FUNCTIONS = {func.__name__: func for func in FUNCTIONS}
RETRY_DELAY = 30


def set_token(token):
    token_path = Path(token)
    if token_path.is_file():
        os.environ['HF_TOKEN'] = token_path.read_text().strip()
    else:
        os.environ['HF_TOKEN'] = token


def load_config(path):
    data = Path(path).read_text()
    data = json.loads(data)
    return data


def do_cache(data):
    for datum in data:
        repo_id = datum['repo_id']
        revision = datum.get('revision', 'main')
        print(f'Caching "{repo_id}" [{revision}]...')
        function = FUNCTIONS[datum.pop('function')]

        while True:
            try:
                path = function(**datum)
                print(f'Cached to "{path}".')
                break
            except Exception as exception:
                print(f'Error caching "{repo_id}": {repr(exception)}. Will retry in {RETRY_DELAY} seconds...')
                time.sleep(RETRY_DELAY)


def run(path_config):
    data = load_config(path_config)
    do_cache(data)


def main():
    parser = argparse.ArgumentParser(description="Cache AI artifacts")
    parser.add_argument('--config', help='Path of config file', required=True)
    parser.add_argument('--token', help='Token for HuggingFace API (can be a path or a string)')

    args = parser.parse_args()

    if args.token:
        set_token(args.token)

    run(args.config)


def tag_model(repo_id: str, tag: str):
    """

    Tag a model repository

    """
    api = huggingface_hub.HfApi()
    return api.create_tag(repo_id, tag=tag, repo_type='model')


def get_hf_cache_path(repo_id, tag=None):
    """

    Get the local cache path for the specified repository and tag.

    """
    tag = tag or 'main'
    path_base = os.path.join(HUGGINGFACE_HUB_CACHE, repo_folder_name(repo_id=repo_id, repo_type='model'))
    ref_path = os.path.join(path_base, "refs", tag)
    if os.path.isfile(ref_path):
        with open(ref_path) as f:
            commit_hash = f.read()
    else:
        raise FileNotFoundError(ref_path)

    path = os.path.join(path_base, "snapshots", commit_hash)
    return path



from datetime import datetime
from pathlib import Path
from setuptools import find_namespace_packages, setup, find_packages

import requirements

ORG, PACKAGE = 'fmtr', 'tools'
ORG_GITHUB = ORG
AUTHOR = 'Frontmatter'
AUTHOR_EMAIL = 'innovative.fowler@mask.pro.fmtr.dev'
DESCRIPTION = 'Collection of high-level tools to simplify everyday development tasks, with a focus on AI/ML'

PATH_BASE = Path(__file__).absolute().parent
IS_SINGLETON = (PATH_BASE / PACKAGE).exists()
VERSION = 'version'

if IS_SINGLETON:
    packages = find_packages()
    package_dir = None
    name = PACKAGE
    path_ver = PATH_BASE / name / VERSION
else:
    packages = find_namespace_packages()
    package_dir = {'': '.'}
    name = f'{ORG}.{PACKAGE}'
    path_ver = PATH_BASE / ORG / PACKAGE / VERSION

__version__ = path_ver.read_text().strip()

setup(
    name=name,
    version=__version__,
    url=f'https://github.com/{ORG_GITHUB}/{name}',
    author=AUTHOR,
    license=f'Copyright © {datetime.now().year} {AUTHOR}. All rights reserved.',
    author_email=AUTHOR_EMAIL,
    description=DESCRIPTION,
    long_description=(PATH_BASE / 'README.md').read_text(),
    long_description_content_type='text/markdown',
    packages=packages,
    package_dir=package_dir,
    package_data={
        name: [VERSION],
    },
    entry_points={
        'console_scripts': requirements.CONSOLE_SCRIPTS,
    },
    install_requires=requirements.INSTALL,
    extras_require=requirements.EXTRAS,
)


import importlib.metadata


def get_version(module):
    """

    Retrieve the version of a specified module.

    """

    if type(module) is not str:
        module = module.__name__
    version = importlib.metadata.version(module)
    return version



def cache_hfh():
    from fmtr import tools
    tools.hfh.main()


def remote_debug_test():
    """

    Test debugger connection

    """
    from fmtr.tools import debug
    debug.trace(is_debug=True)


def shell_debug():
    """

    Test debugger connection

    """
    from fmtr.tools import debug
    debug.debug_shell()
